#![cfg_attr(not(feature = "simulator"), no_std)]
#![cfg_attr(not(feature = "simulator"), no_main)]

mod apps;
mod config;
mod display;
mod energy;
mod fs;
mod gui;
mod keypad;
mod state;
mod strf;
mod text_input;
mod time;

use apps::App;
use display::Display;
use energy::EnergyStatus;
use keypad::KeyEvents;
use state::*;

use arrayvec::ArrayString;
use core::fmt::Write;
use embedded_graphics::{
	mono_font::{
		ascii::{FONT_10X20, FONT_6X10, FONT_9X15},
		MonoTextStyleBuilder,
	},
	pixelcolor::BinaryColor,
	prelude::*,
	primitives::{Line, PrimitiveStyle, PrimitiveStyleBuilder, Rectangle, StrokeAlignment},
	text::{Alignment, Text},
};
#[cfg(not(feature = "simulator"))]
use maduino_zero_4g::{
	self as bsp,
	hal::{
		clock::GenericClockController,
		delay::Delay,
		pac::{CorePeripherals, Peripherals},
		prelude::*,
	},
};
#[cfg(not(feature = "simulator"))]
use panic_halt as _;
use tz::DateTime;

// Check that stack overflow is unlikely
#[cfg(not(feature = "simulator"))]
static_assertions::const_assert!(
	core::mem::size_of::<(
		State,
		ModeState,
		Peripherals,
		maduino_zero_4g::Pins,
		CorePeripherals,
		keypad::Keypad,
		Display,
		Context,
		fs::Fs,
	)>() < 16 * 1024
);

#[cfg_attr(not(feature = "simulator"), bsp::entry)]
fn main() -> ! {
	cfg_if::cfg_if! {
		if #[cfg(not(feature = "simulator"))] {
			let mut peripherals = Peripherals::take().unwrap();
			let mut pins = maduino_zero_4g::Pins::new(peripherals.PORT);
			let core = CorePeripherals::take().unwrap();
			let mut clocks = GenericClockController::with_external_32kosc(
				peripherals.GCLK,
				&mut peripherals.PM,
				&mut peripherals.SYSCTRL,
				&mut peripherals.NVMCTRL,
			);
			let mut delay = Delay::new(core.SYST, &mut clocks);

			let mut fs = fs::Fs::new(atsamd_hal::sercom::spi::Config::new(
				&peripherals.PM,
				peripherals.SERCOM4,
				atsamd_hal::sercom::spi::Pads::default().data_in(pins.sd_miso).data_out(pins.sd_mosi).sclk(pins.sd_sck),
				10_u32.mhz(),
			).enable(), pins.d4.into_push_pull_output());
		}
	}

	let mut state = State {
		energy: EnergyStatus {
			battery: 255,
			charging: false,
		},
		hour: 0,
		minute: 0,
		key_label_left: ArrayString::new_const(),
		key_label_enter: ArrayString::new_const(),
		key_label_right: ArrayString::new_const(),
	};
	let mut mode_state = ModeState::default();

	let mut display = Display::new();
	let mut keypad = keypad::Keypad::default();

	let thin_stroke = PrimitiveStyle::with_stroke(BinaryColor::On, 1);
	let thick_stroke = PrimitiveStyle::with_stroke(BinaryColor::On, 3);
	let fill = PrimitiveStyle::with_fill(BinaryColor::On);
	let statusbar_text_style = MonoTextStyleBuilder::new()
		.font(&FONT_9X15)
		.text_color(BinaryColor::On)
		.background_color(BinaryColor::Off)
		.build();
	let key_label_text_style = MonoTextStyleBuilder::new()
		.font(&FONT_6X10)
		.text_color(BinaryColor::On)
		.build();

	Line::new(Point::new(0, 13), Point::new(199, 13))
		.into_styled(thin_stroke)
		.draw(display.inner_mut())
		.unwrap();
	Line::new(Point::new(0, 189), Point::new(199, 189))
		.into_styled(thin_stroke)
		.draw(display.inner_mut())
		.unwrap();

	#[cfg(feature = "simulator")]
	display.update();

	let mut ctx = Context {
		display: &mut display,
		hm_change: true,
		key_labels_change: false,
		key_events: Default::default(),
		now: DateTime::from_timespec(0, 0, tz::TimeZoneRef::utc()).unwrap(),
		update: true,
		state: &mut state,
	};
	loop {
		let energy_status = energy::get_energy_status();

		if energy_status != ctx.state.energy {
			ctx.update = true;

			Text::with_alignment(
				unsafe { core::str::from_utf8_unchecked(&strf::fmt_energy(&energy_status)) },
				Point::new(0, 9),
				statusbar_text_style,
				Alignment::Left,
			)
			.draw(ctx.display.inner_mut())
			.unwrap();

			ctx.state.energy = energy_status;
		}

		ctx.now = time::now();

		let hour = ctx.now.hour();
		let minute = ctx.now.minute();
		if (hour, minute) != (ctx.state.hour, ctx.state.minute) {
			ctx.update = true;
			ctx.hm_change = true;
			ctx.state.hour = hour;
			ctx.state.minute = minute;

			Text::with_alignment(
				unsafe { core::str::from_utf8_unchecked(&strf::fmt_time_hm(hour, minute)) },
				Point::new(199, 9),
				statusbar_text_style,
				Alignment::Right,
			)
			.draw(ctx.display.inner_mut())
			.unwrap();
		}

		ctx.key_events = keypad.update();

		if let Some(new_mode_state) = match &mut mode_state {
			ModeState::Nothing => Some(ModeState::Clock(Default::default())),
			ModeState::Clock(clock_state) => apps::clock::Clock::update(&mut ctx, clock_state),
			ModeState::Dial(dial_state) => apps::dial::Dial::update(&mut ctx, dial_state),
		} {
			match &mut mode_state {
				ModeState::Nothing => {}
				ModeState::Clock(clock_state) => {
					apps::clock::Clock::on_leave(&mut ctx, clock_state)
				}
				ModeState::Dial(dial_state) => apps::dial::Dial::on_leave(&mut ctx, dial_state),
			}
			mode_state = new_mode_state;
			match &mut mode_state {
				ModeState::Nothing => {}
				ModeState::Clock(clock_state) => {
					apps::clock::Clock::on_enter(&mut ctx, clock_state)
				}
				ModeState::Dial(dial_state) => apps::dial::Dial::on_enter(&mut ctx, dial_state),
			}
		}

		if ctx.key_labels_change {
			let bg_style = PrimitiveStyleBuilder::new()
				.fill_color(BinaryColor::Off)
				.build();
			Rectangle::new(Point::new(0, 190), Size::new(200, 12))
				.into_styled(bg_style)
				.draw(ctx.display.inner_mut())
				.unwrap();

			Text::with_alignment(
				&ctx.state.key_label_left,
				Point::new(0, 197),
				key_label_text_style,
				Alignment::Left,
			)
			.draw(ctx.display.inner_mut())
			.unwrap();
			Text::with_alignment(
				&ctx.state.key_label_right,
				Point::new(200, 197),
				key_label_text_style,
				Alignment::Right,
			)
			.draw(ctx.display.inner_mut())
			.unwrap();
			Text::with_alignment(
				&ctx.state.key_label_enter,
				Point::new(100, 197),
				key_label_text_style,
				Alignment::Center,
			)
			.draw(ctx.display.inner_mut())
			.unwrap();
			ctx.key_labels_change = false;
		}

		#[cfg(feature = "simulator")]
		{
			ctx.display.update();
			std::thread::sleep(core::time::Duration::from_millis(50));
		}
		#[cfg(not(feature = "simulator"))]
		if ctx.update {
			ctx.display.update();
		}
		#[cfg(not(feature = "simulator"))]
		delay.delay_ms(50_u8);

		ctx.update = false;
		ctx.hm_change = false;
	}
}

pub struct Context<'a> {
	pub state: &'a mut State,
	pub display: &'a mut Display,
	pub hm_change: bool,
	pub key_labels_change: bool,
	pub now: DateTime,
	pub update: bool,
	pub key_events: KeyEvents,
}
